/*
Copyright (c) 2011, Intel Corporation
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.
    * Neither the name of Intel Corporation nor the names of its contributors
      may be used to endorse or promote products derived from this software
      without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <grub/extcmd.h>
#include <grub/types.h>
#include <grub/misc.h>
#include <grub/mm.h>
#include <grub/env.h>
#include <grub/err.h>
#include <grub/dl.h>
#include <grub/file.h>
#include <grub/term.h>
#include <grub/time.h>
#include "c.h"
#include "datatype.h"
#include "bitsutil.h"
#include "smp.h"
#include "misc.h"

enum verbose_level {
    V_NONE,
    V_FAIL,
    V_DETAIL,
    V_PASS,
};

static enum verbose_level verbose = V_DETAIL;
static U32 ncpus = 0;
static const CPU_INFO *cpu;
#define ALL_CPUS (~0UL)

struct mem {
    U32 addr;
    U32 value;
};

struct unique_mem {
    U32 count;
    struct mem mem;
};

static struct mem *mem = 0;
static struct unique_mem *unique_mem = 0;

struct msr {
    U32 num;
    U32 status;
    U64 value;
};

static U32 unique_count;

static bool parse_num(const char *str, U64 *result)
{
    bool ret = strtou64_h(str, result);
    grub_errno = GRUB_ERR_NONE;
    return !ret;
}

static U64 eval_var(const char *var)
{
    U64 val;
    char *str = grub_env_get(var);
    if (!str)
        return 0;
    if (parse_num(str, &val))
        return val;
    return 0;
}

static void assign_var(const char *var, U64 val)
{
    char valstr[sizeof("18446744073709551615")];
    grub_snprintf(valstr, sizeof(valstr), "%llu", val);
    grub_env_set(var, valstr);
}

static void passed(void)
{
    assign_var("passed", eval_var("passed") + 1);
}

static void failed(void)
{
    assign_var("failed", eval_var("failed") + 1);
}

static grub_err_t parse_cpu_num(char *str, U32 * out)
{
    U32 num;

    grub_errno = GRUB_ERR_NONE;

    if (grub_strcmp(str, "all") == 0) {
        *out = ALL_CPUS;
        return GRUB_ERR_NONE;
    }

    if (strtou32_h(str, &num) != GRUB_ERR_NONE)
        return grub_errno;

    if (num >= ncpus)
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "CPU number %u invalid", num);

    *out = num;
    return GRUB_ERR_NONE;
}

static U32 viewpoint_env(void)
{
    U32 num;
    char *env = grub_env_get("viewpoint");
    if (!env)
        return ALL_CPUS;
    if (strtou32_h(env, &num) != GRUB_ERR_NONE || num >= ncpus) {
        grub_printf("viewpoint environment variable contains invalid value \"%s\"\n", env);
        return ALL_CPUS;
    }
    return num;
}

static U32 first_cpu(U32 cpu_num)
{
    return cpu_num == ALL_CPUS ? 0 : cpu_num;
}

static U32 last_cpu(U32 cpu_num)
{
    return cpu_num == ALL_CPUS ? ncpus : cpu_num + 1;
}

static void msr_callback(void *param)
{
    struct msr *m = param;
    rdmsr64(m->num, &m->value, &m->status);
}

static void uninit(void)
{
    grub_free(mem);
    grub_free(unique_mem);
}

static grub_err_t oom_uninit(void)
{
    uninit();
    return grub_error(GRUB_ERR_OUT_OF_MEMORY, "Out of memory");
}

static grub_err_t init(void)
{
    uninit();

    ncpus = smp_init();
    if (!ncpus)
        return grub_error(GRUB_ERR_IO, "Failed to initialize SMP");

    cpu = smp_read_cpu_list();
    if (!cpu)
        return grub_error(GRUB_ERR_IO, "Failed to initialize SMP (smp_read_cpu_list)");

    mem = grub_malloc(ncpus * sizeof(*mem));
    if (!mem)
        return oom_uninit();

    unique_mem = grub_malloc(ncpus * sizeof(*unique_mem));
    if (!unique_mem)
        return oom_uninit();

    return GRUB_ERR_NONE;
}

static const char c_help[] =
    "Usage: c \"C-style expression with space-separated tokens\"\n"
    "Evaluate a C expression\n"
    "All evaluation occurs on 64-bit unsigned integers, specified as decimal\n"
    "numbers, hex numbers prefixed by 0x, or named variables taken from the\n"
    "environment.  Variable assignments work, and set variables in the environment.\n"
    "Returns true or false based on the final value of the expression, so\n"
    "\"if c ... \" and \"while c ...\" work.\n"
    "\n"
    "Supported operators, in order from highest to lowest precedence:\n"
    " ()        (parentheses)\n"
    " ! ~ ++ -- (both pre- and post- increment and decrement)\n"
    " * / %\n"
    " + -\n"
    " << >>\n"
    " < <= > >=\n"
    " == !=\n"
    " &         (bitwise and)\n"
    " ^         (bitwise xor)\n"
    " |         (bitwise or)\n"
    " &&        (logical and; WARNING: does not short-circuit)\n"
    " ||        (logical or;  WARNING: does not short-circuit)\n"
    " = += -= *= /= %= <<= >>= &= ^= |=\n"
    " ,         (comma operator)\n";

static grub_err_t grub_cmd_c(grub_command_t cmd, int argc, char **args)
{
    U64 result;
    (void)cmd;

    if (argc == 1 && grub_strcmp(args[0], "--help") == 0) {
        grub_printf("%s", c_help);
        return GRUB_ERR_NONE;
    }

    if (c_expr(argc, args, &result))
        return !result;
    else
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Failed to parse C expression");
}

static U64 adjust_to_nearest(const U64 input, const U64 factor)
{
    U64 val;

    // returns input rounded to the nearest multiple of factor
    // returns (((input + (factor / 2)) / factor) * factor
    val = input + divU64byU64(factor, 2, NULL);
    val = divU64byU64(val, factor, NULL);
    val = mulU64byU64(val, factor, NULL);
    return val;
}

static U64 adjust_to_nearest_fraction(const U64 input, const U64 numerator, const U64 denominator)
{
    U64 val;

    // returns input rounded to the nearest multiple of mixed factor (numerator / denominator)
    // returns ((((input * denominator) + (numerator /2)) / numerator) * numerator) / denominator;
    val = mulU64byU64(input, denominator, NULL);
    val = val + divU64byU64(numerator, 2, NULL);;
    val = divU64byU64(val, numerator, NULL);
    val = mulU64byU64(val, numerator, NULL);
    val = divU64byU64(val, denominator, NULL);

    return val;
}

static void print_hz(U64 hz)
{
    U64 temp64;
    U32 temp32;

    temp64 = grub_divmod64(hz, 1000 * 1000 * 1000, &temp32);
    if (temp64 > 0)
        grub_printf("%llu.%03u GHz", temp64, temp32 / (1000 * 1000));
    else {
        temp64 = grub_divmod64(hz, 1000 * 1000, &temp32);
        if (temp64 > 0)
            grub_printf("%llu.%03u MHz", temp64, temp32 / 1000);
        else
            grub_printf("%llu Hz", hz);
    }
}

static const struct grub_arg_option cpu_freq_options[] = {
#undef OPTION_QUIET
#define OPTION_QUIET 0
    {"quiet", 'q', 0, "No display unless failures occur (default=disabled)", 0, 0},
#undef OPTION_ENV
#define OPTION_ENV 1
    {"env", 'e', 0, "Set environment vars (default=disabled)\n"
                    "\t\t\t\"mperf\"     = IA32_MPERF (Hz)\n"
                    "\t\t\t\"aperf\"     = IA32_APERF (Hz)\n"
                    "\t\t\t\"bclk\"      = BCLK (Hz)\n"
                    "\t\t\t\"adj_mperf\" = *Adjusted IA32_MPERF (Hz)\n"
                    "\t\t\t\"adj_mperf\" = *Adjusted IA32_APERF (Hz)\n"
                    "\t\t\t*Rounded to nearest BCLK multiple", 0, 0},
    {0, 0, 0, 0, 0, 0}
};

#define IA32_MPERF_MSR 0xE7
#define IA32_APERF_MSR 0xE8

static grub_err_t grub_cmd_cpu_freq(struct grub_extcmd *cmd, int argc, char **args)
{
    U32 eax, ecx, dummy;
    U64 mperf, aperf, delta_hz;
    U32 status;
    U64 start, stop;
    U64 mperf_hz, aperf_hz;

    (void)argc;
    (void)args;

    if (init() != GRUB_ERR_NONE)
        return grub_errno;

    cpuid32(0, &eax, &dummy, &dummy, &dummy);
    if (eax < 6) {
        grub_printf("CPUID Leaf 6 is not supported\n");
        return GRUB_ERR_NONE;
    }

    cpuid32(6, &dummy, &dummy, &ecx, &dummy);
    if (!(ecx & (1 << 0))) {
        grub_printf("MPERF/APERF MSRs are not supported\n");
        return GRUB_ERR_NONE;
    }

    wrmsr64(IA32_MPERF_MSR, 0, &status);
    if (status) {
        grub_printf("Writing of IA32_MPERF MSR caused a GPF\n");
        return GRUB_ERR_NONE;
    }
    wrmsr64(IA32_APERF_MSR, 0, &status);
    if (status) {
        grub_printf("Writing of IA32_APERF MSR caused a GPF\n");
        return GRUB_ERR_NONE;
    }
    start = grub_get_time_ms();

    while (grub_get_time_ms() - start < 1000) {
    }

    rdmsr64(IA32_MPERF_MSR, &mperf, &status);
    if (status) {
        grub_printf("Reading of IA32_MPERF MSR caused a GPF\n");
        return GRUB_ERR_NONE;
    }
    rdmsr64(IA32_APERF_MSR, &aperf, &status);
    if (status) {
        grub_printf("Reading of IA32_APERF MSR caused a GPF\n");
        return GRUB_ERR_NONE;
    }
    stop = grub_get_time_ms();

    mperf_hz = grub_divmod64(mperf, stop - start, 0) * 1000;
    aperf_hz = grub_divmod64(aperf, stop - start, 0) * 1000;
    delta_hz = aperf_hz > mperf_hz ? aperf_hz - mperf_hz : mperf_hz - aperf_hz;

    if (!cmd->state[OPTION_QUIET].set) {
        grub_printf("Frequency = ");
        print_hz(mperf_hz);
        grub_printf(" (MPERF)   ");

        print_hz(aperf_hz);
        grub_printf(" (APERF)   ");

        grub_printf("%s", aperf_hz > mperf_hz ? "+" : mperf_hz - aperf_hz ? "-" : "");
        print_hz(delta_hz);
        grub_printf(" (delta)\n");
    }

    if (cmd->state[OPTION_ENV].set) {
        char buf[sizeof("18446744073709551615")];
        U64 bclk, adj_aperf, adj_mperf;

        // Get bclk, convert from MHz to Hz
        bclk = smp_read_bclk() * 1000000;

        // Round to the nearest 100000000 / 12 integer value
        bclk = adjust_to_nearest_fraction(bclk, 100000000, 12);

        // Adjust to nearest multiple of bclk
        adj_aperf = adjust_to_nearest(aperf_hz, bclk);

        // Adjust to nearest 100000 / 12
        adj_aperf = adjust_to_nearest_fraction(adj_aperf, 100000, 12);

        // Adjust to nearest multiple of bclk
        adj_mperf = adjust_to_nearest(mperf_hz, bclk);

        // Adjust to nearest 100000 / 12
        adj_aperf = adjust_to_nearest_fraction(adj_aperf, 100000, 12);

        grub_snprintf(buf, sizeof(buf), "%llu", mperf_hz);
        grub_env_set("mperf", buf);
        grub_snprintf(buf, sizeof(buf), "%llu", aperf_hz);
        grub_env_set("aperf", buf);
        grub_snprintf(buf, sizeof(buf), "%llu", bclk);
        grub_env_set("bclk", buf);
        grub_snprintf(buf, sizeof(buf), "%llu", adj_mperf);
        grub_env_set("adj_mperf", buf);
        grub_snprintf(buf, sizeof(buf), "%llu", adj_aperf);
        grub_env_set("adj_aperf", buf);

    }

    return GRUB_ERR_NONE;
}

#define LATENCY_RECENT_COUNT 6

struct latency_bin {
    const char *desc;
    U64 max;
    U64 avg;
    U64 count;
    U32 recent_index;
    U64 recent_delta[LATENCY_RECENT_COUNT];
    U64 recent_absolute[LATENCY_RECENT_COUNT];
};

#define IA32_TSC_MSR 0x10
/* rdtsc() assumes TSC exists; detect that first by calling rdmsr64 yourself
 * and checking status. */
static U64 rdtsc(void)
{
    U64 temp;
    U32 status;
    rdmsr64(IA32_TSC_MSR, &temp, &status);
    return temp;
}

#define MSR_SMI_COUNT 0x34

static void test_latency(U64 test_duration_tscs, U64 * avg_ptr, U64 * count_ptr, U64 * max_ptr, struct latency_bin *bin, U32 bin_len, bool *smi_count_msr_valid_ptr, U64 *smi_count_msr_delta_ptr)
{
    U64 test_start;
    U64 tsc1, tsc2;
    U64 smi_count1, smi_count2;
    U64 current;
    U32 i, status;
    U64 count = 0, sum = 0, max = 0;
    U32 smi_count_msr_valid = 0;
    struct msr msr;

    msr.num = MSR_SMI_COUNT;
    if (smp_function(cpu[0].apicid, msr_callback, &msr))
        smi_count_msr_valid = !msr.status;

    // For now, assuming that if this MSR exists then it represents the SMI count
    if (smi_count_msr_valid)
        rdmsr64(MSR_SMI_COUNT, &smi_count1, &status);

    for (test_start = tsc1 = tsc2 = rdtsc(); tsc2 - test_start < test_duration_tscs; tsc1 = tsc2, tsc2 = rdtsc()) {
        current = tsc2 - tsc1;

        for (i = 0; i < bin_len; i++)
            if (current <= bin[i].max) {
                bin[i].count++;
                bin[i].avg += current;
                if (bin[i].recent_index != LATENCY_RECENT_COUNT) {
                    bin[i].recent_delta[bin[i].recent_index] = current;
                    bin[i].recent_absolute[bin[i].recent_index] = tsc2;
                    bin[i].recent_index++;
                }
                break;
            }

        if (current > max)
            max = current;
    }

    if (smi_count_msr_valid) {
        rdmsr64(MSR_SMI_COUNT, &smi_count2, &status);
        *smi_count_msr_delta_ptr = smi_count2 - smi_count1;
    }
    *smi_count_msr_valid_ptr = smi_count_msr_valid;

    for (i = 0; i < bin_len; i++) {
        sum += bin[i].avg;
        count += bin[i].count;
        if (bin[i].count)
            bin[i].avg = grub_divmod64(bin[i].avg, bin[i].count, 0);
    }
    *avg_ptr = grub_divmod64(sum, count, 0);
    *count_ptr = count;
    *max_ptr = max;
}

static void print_time(U64 tsc, U64 tsc_per_msec)
{
    U64 temp;
    if ((temp = grub_divmod64(tsc * 1000 * 1000, tsc_per_msec, 0)) < 10 * 1000)
        grub_printf("%4lluns", temp);
    else if ((temp = grub_divmod64(tsc * 1000, tsc_per_msec, 0)) < 10 * 1000)
        grub_printf("%4lluus", temp);
    else if ((temp = grub_divmod64(tsc, tsc_per_msec, 0)) < 10 * 1000)
        grub_printf("%4llums", temp);
    else
        grub_printf("%4llus", grub_divmod64(tsc, 1000 * tsc_per_msec, 0));
}

static const struct grub_arg_option cpu_latency_options[] = {
#undef OPTION_QUIET
#define OPTION_QUIET 0
    {"quiet", 'q', 0, "No display of details (default=disabled)", 0, 0},
#undef OPTION_ENV
#define OPTION_ENV 1
    {"env", 'e', 0, "Set environment vars (default=disabled)\n"
                    "\t\t\t\"max_latency_us\" = Maximum observed latency in microseconds", 0, 0},
    {0, 0, 0, 0, 0, 0},
};
static grub_err_t grub_cmd_cpu_latency(struct grub_extcmd *cmd, int argc, char **args)
{
#define LATENCY_BINS 10
    struct latency_bin bin[LATENCY_BINS];

    bool quiet = false;
    U32 i, j;
    U64 avg, count, max;
    U32 status;
    U64 tsc1, tsc2, start, stop, tsc_per_msec;
    U64 bin_time;
    bool smi_count_msr_valid;
    U64 smi_count_msr_delta = 0;

    (void)argc;
    (void)args;

    quiet = cmd->state[OPTION_QUIET].set;

    if (init() != GRUB_ERR_NONE)
        return grub_errno;

    rdmsr64(IA32_TSC_MSR, &tsc1, &status);
    if (status)
        return grub_error(GRUB_ERR_IO, "Reading of IA32_TSC MSR caused a GPF");

    grub_printf("Warning: touching the keyboard can affect the results of this test.\n");
    grub_printf("Starting pass #1.  Calibrating the TSC.\n");

    start = grub_get_time_ms();
    tsc1 = rdtsc();

    while (grub_get_time_ms() - start < 1000);

    stop = grub_get_time_ms();
    tsc2 = rdtsc();

    tsc_per_msec = grub_divmod64(tsc2 - tsc1, stop - start, 0);
    dprintf("testsuite", "TSC per ms = %llu\n", tsc_per_msec);

    grub_memset(bin, 0, sizeof(bin));

    bin[0].max = grub_divmod64(tsc_per_msec, 1000, 0);
    bin[1].max = grub_divmod64(tsc_per_msec, 100, 0);
    bin[2].max = grub_divmod64(tsc_per_msec, 10, 0);
    bin[3].max = tsc_per_msec;
    for (i = 4; i < LATENCY_BINS; i++)
        bin[i].max = bin[i - 1].max * 10;
    bin[LATENCY_BINS - 1].max = ~0ULL;

    bin[0].desc = "0     < t <=   1us";
    bin[1].desc = "1us   < t <=  10us";
    bin[2].desc = "10us  < t <= 100us";
    bin[3].desc = "100us < t <=   1ms";
    bin[4].desc = "1ms   < t <=  10ms";
    bin[5].desc = "10ms  < t <= 100ms";
    bin[6].desc = "100ms < t <=   1s ";
    bin[7].desc = "1s    < t <=  10s ";
    bin[8].desc = "10s   < t <= 100s ";
    bin[9].desc = "100s  < t         ";

    grub_printf("Starting pass #2.  Wait here, I will be back in 15 seconds.\n");
    test_latency(15 * 1000 * tsc_per_msec, &avg, &count, &max, bin, LATENCY_BINS, &smi_count_msr_valid, &smi_count_msr_delta);

    dprintf("testsuite", "%llu samples; average = %llu\n", count, avg);

    if (!quiet) {
        for (i = 0, bin_time = 1; i < LATENCY_BINS; i++, bin_time *= 10)
            if (bin[i].count) {
                grub_printf("%s; average = ", bin[i].desc);
                print_time(bin[i].avg, tsc_per_msec);
                grub_printf("; count = %llu\n Times between first few observations:", bin[i].count);
                for (j = 0; j < bin[i].recent_index - 1; j++) {
                    grub_printf(" ");
                    print_time(bin[i].recent_absolute[j + 1] - bin[i].recent_absolute[j], tsc_per_msec);
                }
                grub_printf("\n");
            }

        if (smi_count_msr_valid)
            grub_printf("%llu SMI detected using MSR_SMI_COUNT (MSR 0x%x)\n", smi_count_msr_delta, MSR_SMI_COUNT);
    }

    grub_printf("\nSummary of impact: observed maximum latency = ");
    print_time(max, tsc_per_msec);
    grub_printf("\n");

    if (cmd->state[OPTION_ENV].set) {
        char buf[sizeof("18446744073709551615")];
        max = grub_divmod64(max * 1000, tsc_per_msec, 0); /* Convert to microseconds */
        grub_snprintf(buf, sizeof(buf), "%llu", max);
        grub_env_set("max_latency_us", buf);
    }

    return GRUB_ERR_NONE;
}

static void noop_callback(void *param)
{
    (void)param;
}

static const struct grub_arg_option cpu_ping_options[] = {
#undef OPTION_CPU
#define OPTION_CPU 0
    {"cpu", 'c', 0, "CPU number", "CPU", ARG_TYPE_STRING},
    {0, 0, 0, 0, 0, 0}
};

static grub_err_t grub_cmd_cpu_ping(struct grub_extcmd *cmd, int argc, char **args)
{
    U32 i, j;
    U32 cpuNum;
    U32 repeat_count;
    U64 start, stop;
    U32 seconds = 0;

    if (init() != GRUB_ERR_NONE)
        return grub_errno;

    cpuNum = viewpoint_env();

    if (cmd->state[OPTION_CPU].set)
        if (parse_cpu_num(cmd->state[OPTION_CPU].arg, &cpuNum) != GRUB_ERR_NONE)
            return grub_errno;

    if (argc != 1)
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Need 1 argument: repeat_count");
    if (strtou32_h(args[0], &repeat_count) != GRUB_ERR_NONE)
        return grub_errno;

    start = grub_get_time_ms();
    for (j = 0; j < repeat_count; j++) {
        if (grub_checkkey() >= 0 && GRUB_TERM_ASCII_CHAR(grub_getkey()) == GRUB_TERM_ESC)
            break;
        stop = grub_get_time_ms();
        if (stop - start > 1000) {
            start = stop;
            seconds++;
            grub_printf("\r%u second%s (%u%%)", seconds, seconds == 1 ? "" : "s", (j * 100) / repeat_count);
        }
        for (i = first_cpu(cpuNum); i != last_cpu(cpuNum); i++)
            smp_function(cpu[i].apicid, noop_callback, NULL);
    }
    grub_printf("\r");

    return GRUB_ERR_NONE;
}

static const struct grub_arg_option format_options[] = {
#undef OPTION_DEC
#define OPTION_DEC 0
    {"dec", 'd', 0, "Format as decimal (default=enabled)", 0, 0},
#undef OPTION_HEX
#define OPTION_HEX 1
    {"hex", 'x', 0, "Format as hexadecimal", 0, 0},
    {0, 0, 0, 0, 0, 0}
};

static grub_err_t grub_cmd_format(struct grub_extcmd *cmd, int argc, char **args)
{
    int i;
    bool status = true;

    if (!argc)
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Requires format (decimal or hex) & list of environment variables");

    if (cmd->state[OPTION_HEX].set && cmd->state[OPTION_DEC].set)
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "More than one format has been selected");

    for (i = 0; i < argc; i++) {
        char *str = grub_env_get(args[i]);
        U64 num;

        if (strtou64_h(str, &num) == GRUB_ERR_NONE) {
            char buf[sizeof("18446744073709551615")];
            if (cmd->state[OPTION_HEX].set)
                grub_snprintf(buf, sizeof(buf), "0x%llx", num);
            else
                grub_snprintf(buf, sizeof(buf), "%llu", num);
            grub_env_set(args[i], buf);
        } else
            status = false;
    }
    return status ? GRUB_ERR_NONE : grub_error(GRUB_ERR_TEST_FAILURE, "false");
}

static void mem_callback(void *param)
{
    struct mem *m = param;
    m->value = *(U32 *)m->addr;
}

static void get_mem(U32 cpu_num, U32 addr, U64 mask, U32 shift)
{
    U32 i, j;

    unique_count = 0;

    for (i = first_cpu(cpu_num); i != last_cpu(cpu_num); i++) {
        bool found = false;
        mem[i].addr = addr;
        if (!smp_function(cpu[i].apicid, mem_callback, &mem[i]))
            grub_printf("DEBUG: smp_function failed in get_mem\n");
        mem[i].value >>= shift;
        mem[i].value &= mask;

        for (j = 0; j < unique_count; j++)
            if (unique_mem[j].mem.value == mem[i].value) {
                unique_mem[j].count++;
                found = true;
                break;
            }
        if (!found) {
            unique_mem[unique_count].count = 1;
            unique_mem[unique_count].mem = mem[i];
            unique_count++;
        }
    }
}

static const struct grub_arg_option rdmem_options[] = {
#undef OPTION_CPU
#define OPTION_CPU 0
    {"cpu", 'c', 0, "CPU number", "CPU", ARG_TYPE_STRING},
#undef OPTION_VARNAME
#define OPTION_VARNAME 1
    {"varname", 'v', 0, "Save read value into variable VARNAME", "VARNAME", ARG_TYPE_STRING},
#undef OPTION_MASK
#define OPTION_MASK 2
    {"mask", 'm', 0, "Mask to apply to value (default=~0)", "MASK", ARG_TYPE_STRING},
#undef OPTION_SHIFT
#define OPTION_SHIFT 3
    {"shift", 's', 0, "Shift count for mask and value (default=0)", "SHIFT", ARG_TYPE_STRING},
#undef OPTION_QUIET
#define OPTION_QUIET 4
    {"quiet", 'q', 0, "No display unless failures occur (default=disabled)", 0, 0},
    {0, 0, 0, 0, 0, 0}
};

static grub_err_t grub_cmd_rdmem(struct grub_extcmd *cmd, int argc, char **args)
{
    U32 cpuNum;
    U32 memAddr;
    const char *varname;
    U64 mask = ~0ULL;
    U32 shift = 0;
    U32 i, j;
    char mem_desc[128];
    bool status = true;

    if (init() != GRUB_ERR_NONE)
        return grub_errno;

    cpuNum = viewpoint_env();

    if (cmd->state[OPTION_CPU].set)
        if (parse_cpu_num(cmd->state[OPTION_CPU].arg, &cpuNum) != GRUB_ERR_NONE)
            return grub_errno;

    if (cmd->state[OPTION_MASK].set)
        if (strtou64_h(cmd->state[OPTION_MASK].arg, &mask) != GRUB_ERR_NONE)
            return grub_errno;

    if (cmd->state[OPTION_SHIFT].set)
        if (strtou32_h(cmd->state[OPTION_SHIFT].arg, &shift) != GRUB_ERR_NONE)
            return grub_errno;

    if (argc != 1)
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Need 1 argument: memory_address");
    if (strtou32_h(args[0], &memAddr) != GRUB_ERR_NONE)
        return grub_errno;

    if (cmd->state[OPTION_VARNAME].set)
        varname = cmd->state[OPTION_VARNAME].arg;
    else
        varname = NULL;

    get_mem(cpuNum, memAddr, mask, shift);

    // Set fail status for more than one unique value
    if (unique_count != 1)
        status = false;

    if (!cmd->state[OPTION_QUIET].set)
        for (i = 0; i < unique_count; i++) {
            if (mask == ~0ULL) {
                if (shift == 0)
                    grub_snprintf(mem_desc, sizeof(mem_desc), "memory 0x%x", memAddr);
                else
                    grub_snprintf(mem_desc, sizeof(mem_desc), "memory 0x%x shr %u", memAddr, shift);
            } else {
                if (shift == 0)
                    grub_snprintf(mem_desc, sizeof(mem_desc), "memory 0x%x & 0x%llx", memAddr, mask);
                else
                    grub_snprintf(mem_desc, sizeof(mem_desc), "memory 0x%x shr %u & 0x%llx", memAddr, shift, mask);
            }

            grub_printf("%s = 0x%x\n", mem_desc, unique_mem[i].mem.value);

            grub_printf("On %u CPUs:%s", unique_mem[i].count, unique_mem[i].count == ncpus ? " all" : "");
            if (unique_mem[i].count != ncpus)
                for (j = first_cpu(cpuNum); j < last_cpu(cpuNum); j++)
                    if (unique_mem[i].mem.value == mem[j].value)
                        grub_printf(" 0x%x", cpu[j].apicid);
            grub_printf("\n");
        }

    if (varname) {
        char buf[sizeof("0xXXXXXXXXXXXXXXXX")];
        if (unique_count != 1) {
            if (!cmd->state[OPTION_QUIET].set)
                grub_printf("Variable setting requested but memory value not unique\n");
            return grub_error(GRUB_ERR_TEST_FAILURE, "false");
        }
        grub_snprintf(buf, sizeof(buf), "0x%x", unique_mem[0].mem.value);
        grub_env_set(varname, buf);
    }

    return status ? GRUB_ERR_NONE : grub_error(GRUB_ERR_TEST_FAILURE, "false");
}


static const char test_c_help[] =
    "Usage: test_c \"description\" \"C-style expression with space-separated tokens\"\n"
    "Evaluate a C expression and increment pass/fail count based on the result.\n"
    "Refer to \"c --help\" for details on C expression evaluation\n";

static grub_err_t grub_cmd_test_c(grub_command_t cmd, int argc, char **args)
{
    U64 result;
    (void)cmd;

    if (argc == 1 && grub_strcmp(args[0], "--help") == 0) {
        grub_printf("%s", test_c_help);
        return GRUB_ERR_NONE;
    }

    if (c_expr(argc - 1, &args[1], &result)) {
        if (!result) {
            failed();
            if (verbose)
                grub_printf("%s %s\n", args[0], "FAIL");
            return grub_error(GRUB_ERR_TEST_FAILURE, "false");
        } else {
            passed();
            if (verbose >= V_PASS)
                grub_printf("%s %s\n", args[0], "PASS");
            return GRUB_ERR_NONE;
        }
    } else
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Failed to parse C expression");
}

static char *
grub_env_read_test_verbose (struct grub_env_var *var __attribute__ ((unused)), const char *val __attribute__ ((unused)))
{
    // Ignore the string value; return the current value of verbose
    static char buf[sizeof("4294967296")];
    grub_snprintf(buf, sizeof(buf), "%u", verbose);
    return buf;
}

static char *
grub_env_write_test_verbose (struct grub_env_var *var __attribute__ ((unused)), const char *val)
{
    if (strtou32_h(val, &verbose) != GRUB_ERR_NONE)
        return 0;

    return grub_strdup ("");
}

static grub_command_t cmd_c;
static grub_extcmd_t cmd_cpu_freq;
static grub_extcmd_t cmd_cpu_latency;
static grub_extcmd_t cmd_cpu_ping;
static grub_extcmd_t cmd_format;
static grub_extcmd_t cmd_rdmem;
static grub_command_t cmd_test_c;

GRUB_MOD_INIT(testsuite)
{
    /* Make sure $passed and $failed exist. */
    assign_var("passed", eval_var("passed"));
    assign_var("failed", eval_var("failed"));

    // Export these environment variables to make them available in all contexts
    grub_env_export ("test_verbose");
    grub_env_export ("passed");
    grub_env_export ("failed");

    // Register a environment write hook
    grub_register_variable_hook ("test_verbose", grub_env_read_test_verbose, grub_env_write_test_verbose);

  cmd_c = grub_register_command("c", grub_cmd_c, "\"C-style expression with space-separated tokens\"", "Evaluate a C expression.");
  cmd_cpu_freq = grub_register_extcmd("cpu_freq", grub_cmd_cpu_freq, GRUB_COMMAND_FLAG_BOTH,
                                      "[-q] [-e]",
                                      "Compute processor frequency",
                                      cpu_freq_options);
  cmd_cpu_latency = grub_register_extcmd("cpu_latency", grub_cmd_cpu_latency, GRUB_COMMAND_FLAG_BOTH,
                                         "[-q] [-e]",
                                         "Test cpu latency",
                                         cpu_latency_options);
  cmd_format = grub_register_extcmd("format", grub_cmd_format, GRUB_COMMAND_FLAG_BOTH,
                                    "[-d | -x] variable [variable...]",
                                    "Format environment variable",
                                    format_options);
  cmd_cpu_ping = grub_register_extcmd("cpu_ping", grub_cmd_cpu_ping, GRUB_COMMAND_FLAG_BOTH,
                                      "[-c cpu_num] count",
                                      "Ping CPU",
                                      cpu_ping_options);
  cmd_rdmem = grub_register_extcmd("rdmem", grub_cmd_rdmem, GRUB_COMMAND_FLAG_BOTH,
                                   "[-q] [-c cpu_num] [-m mask] [-v VARNAME] [-s shift] memory_addr",
                                   "Read memory",
                                   rdmem_options);
  cmd_test_c = grub_register_command("test_c", grub_cmd_test_c, "\"description\" \"C-style expression with space-separated tokens\"",
                                     "Evaluate a C expression and increment pass/fail count based on the result.");
}

GRUB_MOD_FINI(testsuite)
{
    grub_unregister_command(cmd_test_c);
    grub_unregister_extcmd(cmd_rdmem);
    grub_unregister_extcmd(cmd_format);
    grub_unregister_extcmd(cmd_cpu_ping);
    grub_unregister_extcmd(cmd_cpu_latency);
    grub_unregister_extcmd(cmd_cpu_freq);
    grub_unregister_command(cmd_c);

    // Clear the environment hooks
    grub_register_variable_hook ("test_verbose", 0, 0);

    uninit();
}
